from typing import List, Optional
import time

import fire

from llama import Llama, Dialog, ContextError

import torch
from fairscale.nn.model_parallel.initialize import (
    get_model_parallel_rank,
    destroy_model_parallel
)


def main(
    ckpt_dir: str,
    tokenizer_path: str,
    temperature: float = 0.6,
    top_p: float = 0.9,
    max_seq_len: int = 4096,
    max_batch_size: int = 8,
    max_gen_len: Optional[int] = None,
):
    """
    Entry point of the program for generating text using a pretrained model.

    Args:
        ckpt_dir (str): The directory containing checkpoint files for the pretrained model.
        tokenizer_path (str): The path to the tokenizer model used for text encoding/decoding.
        temperature (float, optional): The temperature value for controlling randomness in generation.
            Defaults to 0.6.
        top_p (float, optional): The top-p sampling parameter for controlling diversity in generation.
            Defaults to 0.9.
        max_seq_len (int, optional): The maximum sequence length for input prompts. Defaults to 4096.
        max_batch_size (int, optional): The maximum batch size for generating sequences. Defaults to 8.
        max_gen_len (int, optional): The maximum length of generated sequences. If None, it will be
            set to the model's max sequence length. Defaults to None.
    """
    # build the model
    generator = Llama.build(
        ckpt_dir=ckpt_dir,
        tokenizer_path=tokenizer_path,
        max_seq_len=max_seq_len,
        max_batch_size=max_batch_size,
    )

    # at this point, we forked into multiple processes, one for each GPU

    dialog: Dialog = []

    # dialog loop
    print("\n===== You can now start typing to chat! =====")
    while True:

        print('\n===== Waiting for user input... =====')

        # only the first process (rank 0) will interact with the user
        user_input = [None]
        if get_model_parallel_rank() == 0:
            user_input = [input("User: ")]

        # broadcast user input to all ranks
        torch.distributed.broadcast_object_list(user_input, src=0)

        print('\n===== Processing query... =====')
        dialog.append({ 'role': 'user', 'content': user_input[0] })

        # run inference
        try:
            start_time = time.time()
            results, ctx = generator.chat_completion(
                [dialog],
                max_gen_len=max_gen_len,
                temperature=temperature,
                top_p=top_p,
            )

            # get the response
            result = results[0]
            role = result['generation']['role']
            content = result['generation']['content']

            # print to user and append to dialog context
            print(f'{role.capitalize()}: {content}')
            dialog.append({ 'role': role, 'content': content })

            print(f'(query took {time.time() - start_time:.2f}s)')

            # context limit reached when generating a sequence
            if ctx >= max_seq_len:
                raise ContextError

        except ContextError:
            # can reach context limit either by the user typing text at the
            # keyboard, or by the model when generating a sequence
            print(f"\n===== Context limit of {max_seq_len} tokens reached! Please restart to keep chatting! Exiting... =====")
            break

    # at the very end, cleanup
    destroy_model_parallel()


if __name__ == "__main__":
    fire.Fire(main)
